/*

  Compiler settings for MSVC 4.2b:

  (suggested)

  preprocessory defs:  NDEBUG ( not  _DEBUG )
  calling convention: fastcall
  processor: pentium
 
   optimizations:
      Assume No Aliasing
      Global Optimizations 
      Generate Intrinsic Functions
      Favor Fast Code 
      Frame-Pointer Omission 
      Full Optimization

  In-line Function Expansion
    Only __inline

  not using precompiled headers
  not incrementally linked or compiled

*/

#include "first.h"
#include "raster_util.h"
#include "clut.h"
#include "raster.h"
#include <assert.h>
#include "rast_i.h"

#pragma optimize("a", on)


//
// statically allocated data
// 

namespace raster
{
 int x_res, y_res;   
 float haze_far_z_inv;
 int bytes_per_pixel;
 int block_bits = 4;

 int span_extents[max_y_res][2];
 int y_min, y_max;

 float plb_sh[3];
 float plb_u[3];
 float plb_v[3];
 float plb_z[3];


 clut_miser::pixel_blend_table_8 pixel_blend_table;

 fill_text_func ** fill_alt;
 fill_text_func ** fill_aft;
 fill_solid_func * fill_als;
 fill_flat_solid_func * fill_afs;


 u16  * frame_buf_starts[max_y_res];
 u32 pixel_blend_mask;

 int   clip_max_x_fx16;
 float clip_max_x_flt;



};


using namespace raster;




//
// external fill function declarations
//


extern "C" fill_text_func fill_alt_16;
extern "C" fill_text_func fill_alt_16_tp;

extern "C" fill_text_func fill_alt_8;
extern "C" fill_text_func fill_alt_8_tp;

extern "C" fill_text_func fill_aft_16;
extern "C" fill_text_func fill_aft_16_tp;

extern "C" fill_text_func fill_aft_8;


extern "C" fill_solid_func fill_als_16;
extern "C" fill_solid_func fill_als_8;


extern fill_flat_solid_func fill_afs_8;
extern fill_flat_solid_func fill_afs_16;




static fill_text_func * fill_alt_16_funcs[2] ={ fill_alt_16, fill_alt_16_tp };
static fill_text_func * fill_alt_8_funcs[2] = { fill_alt_8, fill_alt_8_tp };

static fill_text_func * fill_aft_16_funcs[2] ={ fill_aft_16, fill_aft_16_tp };
static fill_text_func * fill_aft_8_funcs[2] = { fill_aft_8, fill_aft_8 };



//
// function pointers for fill modes
//






static void ensure_intialized();


void raster::set_display_8(const clut_miser::pixel_blend_table_8 * blend_table )
{
  ensure_intialized();

  if( blend_table )
  {
   pixel_blend_table = *blend_table;
  }


  // 8 bit mode
  bytes_per_pixel = 1;
  

 
 //
 // Select 8 bit fill functions
 //
 fill_alt = fill_alt_8_funcs;
 fill_aft = fill_aft_8_funcs;
 fill_als = fill_als_8;
 fill_afs = fill_afs_8;


 refresh_persp_draw_funcs();
}


void raster::set_display_16(const clut_miser::pixel_format_16 & desc )
{
  ensure_intialized();

  //
  // 16 bit
  //
  bytes_per_pixel = 2;
  

  //
  // setup blend mask
  //

  // mask off least significant bits
  pixel_blend_mask = 0;
 
  pixel_blend_mask |=  ((255) >> (desc.r_size_shift+1) ) << (desc.r_place_shift+1); 
  pixel_blend_mask |=  ((255) >> (desc.g_size_shift+1) ) << (desc.g_place_shift+1); 
  pixel_blend_mask |=  ((255) >> (desc.b_size_shift+1) ) << (desc.b_place_shift+1); 

  pixel_blend_mask |= pixel_blend_mask<<16; // double

  pixel_blend_mask |= 1; // last won't underflow


 //
 // Select 16 bit fill functions
 //
 fill_alt = fill_alt_16_funcs;
 fill_aft = fill_aft_16_funcs;
 fill_als = fill_als_16;
 fill_afs = fill_afs_16;
 

 refresh_persp_draw_funcs();
}


/*
void raster::set_display_mode(const display_desc & desc)
{
  ensure_intialized();

  // cache local version
  bytes_per_pixel = desc.bytes_per_pixel;
  


  if( desc.bytes_per_pixel == 2 )
  {
   // mask off least significant bits
    pixel_blend_mask = 0;
 
    pixel_blend_mask |=  ((255) >> (desc.r_size_shift+1) ) << (desc.r_place_shift+1); 
    pixel_blend_mask |=  ((255) >> (desc.g_size_shift+1) ) << (desc.g_place_shift+1); 
    pixel_blend_mask |=  ((255) >> (desc.b_size_shift+1) ) << (desc.b_place_shift+1); 

    pixel_blend_mask |= pixel_blend_mask<<16; // double

    pixel_blend_mask |= 1; // last won't underflow
  }



 //
 // 8 bit?
 //
 if( bytes_per_pixel == 1 )
 {
  fill_alt = fill_alt_8_funcs;
  fill_aft = fill_aft_8_funcs;
  fill_als = fill_als_8;
 }
 // 
 // 16 bit
 //
 else
 {
  fill_alt = fill_alt_16_funcs;
  fill_aft = fill_aft_16_funcs;
  fill_als = fill_als_16;
 }

 refresh_persp_draw_funcs();
}
*/

void raster::set_persp_block_dim(int dim )
{

  if( dim == 4 )
    block_bits = 2;
  else if( dim == 8 )
   block_bits = 3;
  else if( dim == 16 )
   block_bits = 4;
  else
    assert( 0 ); // bad dimension, only suports 4,8,16

 refresh_persp_draw_funcs();
}



// 
// table of floats to avoid fild 2 cycle latency
//
float rast_int_to_flt[max_x_res];


//
// unprotect asm for modifying
//
#if 0
// -qq- first.h includes windows.h which defines this for us
extern "C" __stdcall VirtualProtect(void*,u32,u32,u32*);
#endif

//  needs disprotection
extern "C" int self_modify_a_start,self_modify_a_end;
extern "C" int self_modify_p_start,self_modify_p_end;


#define PAGE_EXECUTE_READWRITE 0x40     

// performs one time initialization routines
static void ensure_intialized()
{
 // already initialized?
 if( rast_int_to_flt[1] != 0 )
   return;


  //
  // allow use of source code variables
  //

  //
  // in a_rast.asm
  //
  u32 old_protection =0;
  if(!VirtualProtect( &self_modify_a_start, (int)&self_modify_a_end - (int)&self_modify_a_start,
                      PAGE_EXECUTE_READWRITE,&old_protection) )
   {
    // must be able to unprotect code
    assert( 0 );
   }
  //
  // in p_rast.asm
  //
  if(!VirtualProtect( &self_modify_p_start, (int)&self_modify_p_end - (int)&self_modify_p_start,
                      PAGE_EXECUTE_READWRITE,&old_protection) )
   {
    // must be able to unprotect code
    assert( 0 );
   }


  for(int x=0;x<max_x_res;++x)
   {
    rast_int_to_flt[x] = x;
   }
}














//
// transparency indexes
//



extern "C" byte fill_alt_16_tp_index[], 
                fill_alt_8_tp_index[], 
                fill_aft_16_tp_index[], 
                fill_aft_8_tp_index[],
                tp_index_8_02[],
                tp_index_8_03[],
                tp_index_8_04[],
                tp_index_8_12[],
                tp_index_8_13[],
                tp_index_8_14[],
                tp_index_16_02[],
                tp_index_16_03[],
                tp_index_16_04[],
                tp_index_16_12[],
                tp_index_16_13[],
                tp_index_16_14[];

void raster::set_transparency_index(byte index)
{
 fill_alt_16_tp_index[-1]=  index;
 fill_alt_8_tp_index[-1]=  index;
 fill_aft_16_tp_index[-1]=  index;
 fill_aft_8_tp_index[-1]=  index;
 tp_index_8_02[-1] = index;
 tp_index_8_12[-1] = index;
 tp_index_8_03[-1] = index;
 tp_index_8_13[-1] = index;
 tp_index_8_04[-1] = index;
 tp_index_8_14[-1] = index;

 tp_index_16_02[-1] = index;
 tp_index_16_12[-1] = index;
 tp_index_16_03[-1] = index;
 tp_index_16_13[-1] = index;
 tp_index_16_04[-1] = index;
 tp_index_16_14[-1] = index;
}


void raster::set_surface(int _x_res, int _y_res, void * buffer, u32 buf_pitch)
{
 assert( _x_res <= max_x_res );
 assert( _y_res <= max_y_res );

 static u32 cur_buf_pitch=0;

 // do we need to remake the offsets?
 if(( _y_res != y_res ) || ( buffer != frame_buf_starts[0] ) || ( cur_buf_pitch != buf_pitch ) )
 {
   cur_buf_pitch = buf_pitch;

   for(int y=0;y<_y_res;++y, (int &)buffer += buf_pitch)
   {
    frame_buf_starts[y] = (u16*)buffer;
   }
 }
 x_res = _x_res;
 y_res = _y_res;

 clip_max_x_fx16 = x_res * FX1.i() + FX1.i()/2;
 clip_max_x_flt = x_res -0.5;
}
